/**********************************************************
 * File: KDTree.cpp
 * Author: TODO: Your name here!
 *
 * TODO: Describe your implementation here.
 */

#include "KDTree.h"
#include <iostream>
#include <cmath>

/****** Module helper functions ******/

/* Computes the Euclidean distance between two vectors. */
static double Distance(const vector<double>& one, const vector<double>& two)
{
    double result = 0.0;
    for (size_t i = 0; i < one.size(); ++i)
        result += (one[i] - two[i]) * (one[i] - two[i]);
    return result;
}

/****** Implementation of const_iterator ******/

/* Default constructor says that the current node is NULL
 * and doesn't populate the frontier.
 */
KDTree::const_iterator::const_iterator()
{
    curr = NULL;
}

/* The private constructor for the const_iterator type needs to do
 * two things:
 *
 * 1. Set the iterator to start at the indicated location.
 * 2. Build the frontier of nodes yet to visit.
 *
 * This second step is tricky and most of what we need to do
 * concerns that step.  The basic idea is to walk up to the
 * root of the tree, adding to the frontier any nodes that
 * are to the right of the current node (nodes to the left
 * would already have been visited).
 */
KDTree::const_iterator::const_iterator(const KDTreeNode* where)
{
    curr = where;

    /* If this is a null iterator, we're done. */
    if (!where) return;

    /* Add the children of the current node to the frontier. */
    if (where->right)
        toExplore.push_front(where->right);
    if (where->left)
        toExplore.push_front(where->left);

    /* Walk up the tree adding the other children to the explore list.  We will
     * only add nodes to the right.
     */
    for (const KDTreeNode* parent = where->parent; parent != NULL; where = parent, parent = parent->parent)
    {
        /* See where we came from.  If we came from a left child, add the
         * right child.  Otherwise do nothing.
         */
        if (parent->left == where && parent->right)
            toExplore.push_back(parent->right);
    }
}

/* Two iterators are equal if they're considering the same point.  This
 * automatically implies that the frontiers are equal.
 */
bool KDTree::const_iterator::operator== (const_iterator other) const
{
    return curr == other.curr;
}

/* Disequality is implemented in terms of equality. */
bool KDTree::const_iterator::operator!= (const_iterator other) const
{
    return !(*this == other);
}

/* To advance the iterator to the next location, we explore the next element
 * in the frontier and then update the frontier to consider that location's
 * children.
 */
KDTree::const_iterator& KDTree::const_iterator::operator++ ()
{
    /* If no next location exists, we're done. */
    if (toExplore.empty())
    {
        curr = NULL;
        return *this;
    }

    /* Otherwise, move to the next location. */
    curr = toExplore.front();
    toExplore.pop_front();

    /* Add the left and right children (right child first) */
    if (curr->right)
        toExplore.push_front(curr->right);
    if (curr->left)
        toExplore.push_front(curr->left);

    return *this;
}

/* Postfix increment implemented in terms of prefix increment. */
const KDTree::const_iterator KDTree::const_iterator::operator++ (int)
{
    const_iterator result = *this;
    ++*this;
    return result;
}

/* Derefencing the iterator simply returns a const reference to the
 * data at the current node.
 */
const KDTree::value_type& KDTree::const_iterator::operator*  () const
{
    return curr->value;
}

/* Arrow implemented in terms of *. */
const KDTree::value_type* KDTree::const_iterator::operator-> () const
{
    return &**this;
}

/****** Implementation of KDTreeNode ******/

/* Main constructor initializes the location's value and parent, 
 * then sets the child pointers to NULL.
 */
KDTree::KDTreeNode::KDTreeNode(const value_type& value, KDTreeNode* parent) :
   value(value), parent(parent)
{
    left = right = NULL;
}

/* Destructor recursively deletes the children of the current node. */
KDTree::KDTreeNode::~KDTreeNode()
{
    /* Note that there are no NULL-checks here.  In C++, calling
     * delete on a NULL pointer is safe and has no effect.
     */
    delete left;
    delete right;
}

/****** Implementation of KDTree ******/

/* Comparison functions used in the constructor.  You should not need to modify
 * or use these functions.
 */
bool KDTree::ValuesHaveEqualPoints(const value_type* one, const value_type* two)
{
        return one->first == two->first;
}
bool KDTree::CompareValueTypesIndirectly(const value_type* one, const value_type* two)
{
    return *one < *two;
}

/* KDTree destructor needs to clean up resources. */
KDTree::~KDTree()
{
    /* Feel free to modify this function when writing the copy constructor
     * and assignment operator.
     */
    delete root;
}

/* Type: AxisCompare
 * --------------------------------------------------------
 * A comparison functor that decides how two value_types
 * compare to one another along some axis.  The constructor
 * of this AxisCompare functor should take in axis index
 * (i.e. which component of the vectors to compare) and its
 * operator() should take in two const KDTree::value_type*s
 * and how the indicated components of the vectors pointed at
 * compare.
 */
class AxisCompare
{
public:
    /* Constructor: AxisCompare(size_t n);
     * Usage: AxisCompare axis(0); // Compares the first component of each point.
     * --------------------------------------------------------
     * Constructs a new AxisCompare functor that compares points
     * along axis n.
     */
    explicit AxisCompare(size_t axis);

    /* Function: bool operator() (const KDTree::value_type* one,
     *                            const KDTree::value_type* two);
     * --------------------------------------------------------
     * Given two pointers to value_types (which are
     * pair<vector<double>, string>s), and remembering the n
     * parameter specified in the constructor, returns whether
     * the first parameter's vector's nth element is less than
     * the second parameter's vector's nth element.
     */
    bool operator()(const KDTree::value_type* one, const KDTree::value_type* two) const;
    
private:
    // TODO: Private data members? //
};

AxisCompare::AxisCompare(size_t axis)
{
    // TODO: Implement this constructor.  As a hint,
    // it should be possible to do this in one line of code.
}

bool AxisCompare::operator() (const KDTree::value_type* one, const KDTree::value_type* two) const
{
    // TODO: Implement this function.  As a hint,
    // it should be possible to do this in one line of code.
    return false;
}

/* Given a range of iterators delineating key/value pairs, constructs a kd-tree
 * out of the data using axis as the first splitting axis and parent as the
 * parent of the tree.
 */
KDTree::KDTreeNode* KDTree::recMakeKDTree(vector<value_type*>::iterator begin,
                                          vector<value_type*>::iterator end,
                                          size_t axis,
                                          KDTreeNode* parent)
{
    /* If there is no data to build into a tree, we hand back the NULL
     * tree.
     */
    if (begin == end) return NULL;
    
    /* Get an iterator to the middle of the iterator range. */
    vector<value_type*>::iterator mid = begin + distance(begin, end) / 2;

    /* Partition the data so that the median of the points is in the middle.
     * This uses the AxisCompare functor that you will be writing to do the
     * partioning.
     *
     * An important detail is that multiple points from [begin, end) might
     * have the same axis'th component.  Consequently, after we compute the
     * median, we have to do a bit more processing to check what element will
     * actually be used as the partition.
     */
    nth_element(begin, mid, end, AxisCompare(axis));

    /* Now that we know what value goes in the middle, we need to count how
     * many elements are less than, equal to, and greater than the median.
     * This will influence where we put the partition point.
     */
    int less = 0, equal = 0, greater = 0;
    AxisCompare comp(axis);
    for (vector<value_type*>::iterator itr = begin; itr != end; ++itr)
    {
        /* A value less than the median. */
        if (comp(*itr, *mid)) ++less;
        
        /* A value greater than the median. */
        else if (comp(*mid, *itr)) ++greater;
        
        /* Equal to the median. */
        else ++equal;
    }

    /* We'll consider two options: either partition across the median element, or
     * partition across the element that follows the median.  This depends on how
     * good of a split we can get.  This code determines the relative sizes of the
     * splits and picks the better of the two.
     */
    const int totalElems = less + equal + greater;
    const int splitOne   = less + equal;
    const int splitTwo   = equal + greater;

    /* Case one: Choose first elem above the median. */
    if (abs(splitOne - totalElems / 2) < abs(splitTwo - totalElems / 2))
        mid = begin + less + equal;
    /* Case two: Choose first element having median value. */
    else
        mid = begin + less;

    /* Use nth_element to reorder everything correctly one more time. */
    nth_element(begin, mid, end, comp);

    /* Construct a new tree node that holds the median data and has the
     * specified parent.
     */
    KDTreeNode* result = new KDTreeNode(**mid, parent);
    
    /* Recursively construct trees for all other data points, using the next axis as
     * the splitting axis and the new node as the parent.
     */
    result->left  = recMakeKDTree(begin,   mid, (axis + 1) % dimension(), result);
    result->right = recMakeKDTree(mid + 1, end, (axis + 1) % dimension(), result);
    return result;
}

/* Returns the number of dimensions in the tree. */
size_t KDTree::dimension() const
{
    return numDimensions;
}

/* Build a const_iterator starting at the root of the tree.  This will explore the
 * entire tree and thus will visit every element.
 */
KDTree::const_iterator KDTree::begin() const
{
    return const_iterator(root);
}

/* Building a const_iterator out of NULL signifies that there is no more input. */
KDTree::const_iterator KDTree::end() const
{
    return const_iterator(NULL);
}

size_t KDTree::size() const
{
    // TODO: Implement this function!
    return 0;
}

bool KDTree::empty() const
{
    // TODO: Implement this function!
    return false;
}

size_t KDTree::count(const vector<double>& v) const
{
    // TODO: Implement this function!
    return 0;
}

KDTree::const_iterator KDTree::find(const vector<double>& key) const
{
    // TODO: Implement this function!
    return end();
}

string KDTree::kNearestNeighborValue(const vector<double>& v, size_t k) const
{
    // TODO: Implement this function!
    return "";
}
